#include "../include/file.h"

// TO DO: virer a, et utiliser kalloc
#include <stdlib.h>


/***************************************/
/* Dfinitions                         */
/***************************************/
#define NB_FILES 20 // Nombre max de fichiers ouverts simultanment



/***************************************/
/* Structure vortex                    */
/***************************************/
typedef struct Vortex
{
  // Attributs gnraux
  int typefichier;
  int ID;

  // Secteur actuel
  unsigned char secteur[512];
  int CurSecteur;
  int Current; // Position actuelle dans le secteur
  int Taille; // Taille du fichier
} Vortex;


/*****************************************/
/* Variables d'tat                      */
/*****************************************/

// Type fichier:
// 0 -> rootdir
// 1 -> fichier normal
// 2 -> directory
unsigned char secteur [512];
Vortex * f[NB_FILES];

// FAT
char * Fat;
int IsFatLoaded;

/****************************************/
/* Interface des fonctions principales  */
/****************************************/
int fakemount(char);
int fakeopenf(char *);
int fakereadf(int, char *, int);
void fakeclosef(int);

/****************************************/
/* Interface des fonctions externes     */
/****************************************/
void initfloppy();
void readsector(int, char *);


/****************************************/
/* Interface des fonctions secondaires  */
/****************************************/
int IsFat12();
void LoadFat(int num);
int GetNextEntry(Vortex * f, unsigned char * entry);
int  ProcessEntry(unsigned char * entry, fichier12 * f12);
void LoadSecteur(Vortex * f, int num);
void CopieNom(char * unin, char * nom);
void getheure(heure * h, unsigned short hh);
void getdate(date * d, unsigned short dd);
int recherchef12(char * s, int numfile, fichier12 * f12);
unsigned short GetNextSecteur(unsigned short secteur);

// Recherche dans la liste des fichiers
int ChercheNumLibre();

/*****************************************/
/* Interface des fonctions utilitaires   */
/*****************************************/
void memcopy(void * dest, void * src, int size);
int strncomp(char * str1, char * str2, int n);
int getnextchar(char c, char * s);
int getprecchar(char c, char * s);
int stringlen(char * s);

/******************************************/
/* Implmentation de la fat12             */
/******************************************/

int fakemount(char mode)
{
  // On monte le truc
  if(mode == 0)
    {
      int i;
      initfloppy();
      Fat = (char *)malloc(512*9*sizeof(char));
      LoadFat(1);
      IsFatLoaded = 1;
      // On vide la structure de fichiers
      for(i = 0; i < NB_FILES; i++)
	{
	  f[i] = NULL;
	}
    }
  else
    {
      int i;
      // On dmonte le truc
      free(Fat);
      Fat = NULL;
      IsFatLoaded = -1;
      // On libre tous les fichiers ouverts
      for(i = 0; i < NB_FILES; i++)
	{
	  if(f[i] != NULL)
	    {
	      fakeclosef(i);
	      free(f[i]);
	      f[i] = NULL;
	    }
	}
    }
    return 0;
}

int fakeopenf(char * nom)
{
  int num;
  // Est-ce que la disquette est monte???
  if(IsFatLoaded != 1)
    {
      return -1;
    }

  // On cr un vortex	
  // Recherche du prochain numro de libre
  num = ChercheNumLibre();
  if(num == -1)
    {
      return -4;
    }
  f[num] = (Vortex *)malloc(sizeof(Vortex)); 
  f[num]->typefichier = -1;
  f[num]->ID = -1;

  // On vrifie que la disquette est bien fat12 (et par extension, qu'elle est toujours prsente)
  if(!IsFat12())
    {
      free(f[num]);
      f[num] = NULL;
      return -2;
    }
  
  if(strncomp(nom, "/", 2))
    {
      f[num]->typefichier = 0;
      f[num]->CurSecteur = -1;
      f[num]->Current = -1;
    }
  else
    {
      // On choppe le / prcdent
      int c = getprecchar('/', nom);
      int cf = c;
      char recnom[255];
      int numfile;
      fichier12 f12;
      // On copie le nom du dessous
      memcopy(recnom, nom, c*sizeof(char));
      // Ajustements sur le nom
      if(c > 0)
	{
	  c--;
	  while((recnom[c] == '/')&&(c > 0)) c--;
	  recnom[c+1] = '\0';
	}
      if(c == 0)
	{
	  // On est  la fin
	  recnom[0] = '/';
	  recnom[1] = '\0';
	}
      // On ouvre le fichier prcdent
      numfile = fakeopenf(recnom);
      if(numfile < 0) return numfile;
      if(recherchef12(&nom[cf+1], numfile, &f12) == -1) return -3;
      fakeclosef(numfile);
      // On regarde les attributs du nouveau fichier, et on l'ouvre
      if(f12.attributs & 0x10) // C'est un sous-rpertoire
	{
	  f[num]->typefichier = 2;
	  f[num]->CurSecteur = 31 + f12.cluster;
	  // Chargement du secteur correspondant au cluster
	  LoadSecteur(f[num], f[num]->CurSecteur);
	  f[num]->Current = 0;
	  return(f[num]->ID = num);
	}
      else
	{
	  // C'est un fichier normal. Il nous faut sa taille
	  f[num]->typefichier = 1;
	  f[num]->CurSecteur = 31 + f12.cluster;
	  // Chargement du secteur correspondant au cluster
	  LoadSecteur(f[num], f[num]->CurSecteur);
	  f[num]->Current = 0;
	  f[num]->Taille = f12.taille;
	  return(f[num]->ID = num);
	}
    }
  
  return (f[num]->ID = num);
}

int fakereadf(int nf, char * data, int nb)
{
  // On vrifie que le numro est valide
  if((nf < 0) || (nf >= NB_FILES)) return -1;
  if(f[nf] == NULL) return -1;
  // On regarde les attributs du fichier
  if((f[nf]->typefichier == 0)||(f[nf]->typefichier == 2))
    {
      // C'est une directory. On choppe la prochaine entre
      unsigned char entry[32];
      ((fichier12*)data)->lfn = 0;
      do
	{
	  GetNextEntry(f[nf], entry);
	}
      while(ProcessEntry(entry, (fichier12 *)data) == 0);
      if(((fichier12 *)data)->lfn == -1) return -1;
      return 1;
    }
  else
    {
      char * c = data; // Pointeur sur le buffer d'criture
      int nbr = 0; // Nombre rel d'octets lus

      if(f[nf]->Current + nb > f[nf]->Taille)
	{
	  nb = f[nf]->Taille - f[nf]->Current;
	}

      while(nbr < nb)
	{
	  // Combien faut-il lire?
	  int NbALire = 512 - (f[nf]->Current%512);
	  if(NbALire > nb - nbr)
	    {
	      NbALire = nb - nbr;
	    }
	  memcopy(c, f[nf]->secteur, NbALire * sizeof(char));
	  // On passe au prochain secteur
	  f[nf]->CurSecteur = GetNextSecteur(f[nf]->CurSecteur-31)+31;
	  f[nf]->Current += NbALire;
	  nbr += NbALire;
	  c += NbALire;
	  if(f[nf]->CurSecteur >= 0xff8) return nbr;
	  LoadSecteur(f[nf], f[nf]->CurSecteur);
	}
      return nbr;
    }
}

void fakeclosef(int nf)
{
  // On vrifie que le numro est valide
  if((nf < 0) || (nf >= NB_FILES)) return;
  if(f[nf] == NULL) return;
  // TODO: si une modification a eut lieu, copier le secteur sur la disquette
  // On libre tout
  free(f[nf]);
  f[nf] = NULL;
}

/*****************************************/
/* Implmentation des fcts secondaires   */
/*****************************************/
void LoadSecteur(Vortex * f, int num)
{
  readsector(num, f->secteur);
  f->CurSecteur = num;
}

void LoadFat(int num)
{
  int depart;
  int i;
  if(num == 1) depart = 1;
  else depart = 10;

  // On copie tous les secteurs de la Fat 1  10.
  // Attention! Vu que c'est du DMA, il faut passer par notre secteur
  // temporaire en variable globale (sous la barre des 1 mega)
  for(i = depart; i < depart + 9; i++)
    {
      readsector(i, secteur);
      memcopy((Fat+(i-depart)*512), secteur, 512*sizeof(char));
    }
}

int IsFat12()
{
  char type[8];
  // Puis on lit le premier secteur
  readsector(0, secteur);
  // Et on regarde si c'est bien de la fat12
  memcopy(type, &secteur[0x36], 8*sizeof(char));
  if(strncomp(type, "FAT12   ", 8))
    {
      return 0 ;
    }
  return 1; 
}

int GetNextEntry(Vortex * f, unsigned char * entry)
{
  // Est-ce la rootdir?
  if(f->typefichier == 0)
    {
      // si rien n'est charg, on charge le rootsecteur
      if(f->CurSecteur < 0)
	{
	  LoadSecteur(f, 19);
	  f->Current = 0;
	}
      // Si Current est au max, on passe au suivant
      if(f->Current >= 512)
	{
	  f->CurSecteur++;
	  LoadSecteur(f, f->CurSecteur);
	  f->Current = 0;
	}
      // On lit l'entre courante
      memcopy(entry, f->secteur + f->Current, 32*sizeof(char));
      // Puis on incrmente Current
      f->Current += 32;
    }
  else
    {
      // C'est une subdirectory
      // Quelques tests...
      if(f->CurSecteur < 0) return -1; // Gros problme...
      // Si Current est au max, on passe au suivant
      if(f->Current >= 512)
	{
	  f->CurSecteur = GetNextSecteur(f->CurSecteur - 31) + 31;
	  LoadSecteur(f, f->CurSecteur);
	  f->Current = 0;
	}
      // On lit l'entre courante
      memcopy(entry, f->secteur + f->Current, 32*sizeof(char));
      // Puis on incrmente Current
      f->Current += 32;
    }
  return 0;
}

int ProcessEntry(unsigned char * entry, fichier12 * f12)
{
  if(entry[0] == 0) // C'est fini!!
    {
      f12->lfn = -1;
      return -1;
    }
  else if(entry[0] == 0xe5)
    {
      // Cherchons l'entre suivante
      return 0;
    }
  else 
    {
      unsigned short t;
      unsigned short cluster;
      // C'est une vraie entre!! Youpie!
      // Testons s'il s'agit d'une LFN:
      memcpy(&cluster, &entry[0x1a], 2*sizeof(char));
      if((entry[0x0b] == 0x0f)&&(cluster == 0))
	{
	  // C'est une LFN!!
	  int n = 13 * ((entry[0]&0x0f) - 1);
	  if(entry[0]&0x40) // C'est la premire lfn... Argh!!
	    {
	      // Nettoyage de tout le bousin
	      int i;
	      int n = 13 * ((entry[0]&0x0f) - 1);
	      for(i=0; i < 255; i++)
		{
		  f12->nom[i] = ' ';
		}
	      // Passage en lfn
	      f12->lfn = 1;
	      f12->nom[n + 13] = '\0';
	      f12->ext[0] = '\0';
	    }
	  // Puis copie du nom
	  CopieNom(&entry[0], &f12->nom[n]);
	  return 0;
	}
      else
	{
	  // Etait-on en fichier long? Si oui, on ne copie pas le nom, ni l'extension
	  if(f12->lfn == 0)
	    {
	      memcpy(f12->nom, &entry[0], 8*sizeof(char));
	      memcpy(f12->ext, &entry[8], 3*sizeof(char));
	      f12->nom[8] = '\0';
	      f12->ext[3] = '\0';
	    }
	  memcpy(&t, &entry[0xe], 2*sizeof(char));
	  getheure(&f12->heuremodif, t);
	  memcpy(&t, &entry[0x10], 2*sizeof(char));
	  getdate(&f12->datemodif, t);
	  memcpy(&t, &entry[0x12], 2*sizeof(char));
	  getdate(&f12->dernieraccess, t);
	  memcpy(&t, &entry[0x16], 2*sizeof(char));
	  getheure(&f12->heurecreation, t);
	  memcpy(&t, &entry[0x18], 2*sizeof(char));
	  getdate(&f12->datecreation, t);
	  memcpy(&f12->taille, &entry[0x1c], 4*sizeof(char));
	  memcpy(&f12->attributs, &entry[0x0b], sizeof(char));
	  memcpy(&f12->cluster, &entry[0x1a], 2*sizeof(char));
	  return 1;
	}
    }
}

void getheure(heure * h, unsigned short hh)
{
  h->heures = (char)((hh & 0xf800)>>11);
  h->minutes = (char)((hh & 0x07e0)>>5);
  h->secondes = (char)((hh & 0x001f)*2);
}

void getdate(date * d, unsigned short dd)
{
  d->annee = 1980 + ((dd & 0xfe00)>>9);
  d->mois = (char)((dd & 0x01e0)>>5);
  d->jour = (char)(dd & 0x001f);
}

int ChercheNumLibre()
{
  int i;
  // On commence  3: en effet, on rserve 0, 1, 2 pour un usage ultrieur
  for(i = 3; i < NB_FILES; i++)
    {
      if(f[i] == NULL) return i;
    }
  return -1;
}

void CopieNom(char * unin, char * nom)
{
  int i;
  for(i = 0; i < 5; i++)
    {
      nom[i] = unin[i*2 + 1];
    }
  for(i = 7; i < 13; i++)
    {
      nom[i-2] = unin[i*2];
    }
  for(i = 14; i < 16; i++)
    {
      nom[i-3] = unin[i*2];
    }
}

int recherchef12(char * s, int numfile, fichier12 * f12)
{
  while(fakereadf(numfile, (char *)f12, 1) >= 1)
    {
      char nom[255];
      // Il faut recrer le nom. Si c'est un long LFN, piece of cake. Sinon, faut recoller les bouts
      if(f12->lfn == 1)
	{
	  int n = strlen(f12->nom);
	  memcpy(nom, f12->nom, n+1);
	  while(nom[n-1] == ' ') n--;
	  nom[n] = '\0';
	}
      else
	{
	  int n = strlen(f12->nom);
	  memcopy(nom, f12->nom, n);
	  while(nom[n-1] == ' ') n--;
	  if(f12->attributs & 0x10)
	    {
	      nom[n] = '\0';
	    }
	  else
	    {
	      nom[n] = '.';
	      memcopy(&nom[n+1], f12->ext, 4);
	    }
	}
      // Maintenant, on peut comparer les noms
      if(strncomp(s, nom, 255) == 1) return 1;
    }
  return -1;
}

unsigned short GetNextSecteur(unsigned short secteur)
{
  unsigned short t;
  if(secteur % 2 == 0)
    {
      memcopy(&t, &Fat[(secteur * 3) / 2], 2*sizeof(char));
      t = t & 0x0fff;
    }
  else
    {
      memcopy(&t, &Fat[(secteur * 3) / 2], 2*sizeof(char));
      t = t >> 4;
    }
  return t;
}

/******************************************/
/* Implmentation des utilitaires         */
/******************************************/

void memcopy(void * dest, void * src, int size)
{
  char * pdest = (char *)dest;
  char * psrc = (char *)src;
  int i;
  for(i = 0; i < size; i++, pdest++, psrc++)
    {
      *pdest = *psrc;
    }
}

int strncomp(char * str1, char * str2, int n)
{
  int i;
  for(i = 0; i < n; i++, str1++, str2++)
    {
      if(*str1 != *str2) return 0;
      if(*str1 == '\0') return 1;
    }
  return 0;
}

// La fonction getnextchar renvoie le dernier caractre c de la premire suite de caractres c rencontrs
// exemple: getnextchar('c', "abcccdefabcdef" renvoie 4, car 4 est le dernier c de la premire suite de c
// Si c n'est pas trouv, la fonction renvoie -1
int getnextchar(char c, char * s)
{
  int i = 0;
  do
    {
      if(s[i] == c)
	{
	  // On entre dans la deuxime phase: aller au fond de la suite de c
	  while(s[i] == c) i++;
	  return i;
	}
      i++;
    }
  while(s[i] != '\0');
  return -1;
}

// Fonctionne grosso modo comme getnextchar, mais  l'envers (eh oui!)
int getprecchar(char c, char * s)
{
  int i = stringlen(s);
  do
    {
      if(s[i] == c) return i;
      i--;
    }
  while(i >= 0);
  return -1;
}

int stringlen(char * s)
{
  int i = 0;
  while(s[i] != '\0') i++;
  return i;
}

