#include "kalloc.h"
#include "mem.h"

// Implantation du heap
#define START_ADDR 0x20000000
#define LIMIT 10000000
#define FREE 0x1
#define NULL 0x0

typedef struct memheader
{
  struct memheader * previous;
  struct memheader * next;
  int state;
} memheader;

void * kalloc(int size)
{
  // D'abord, on va charcher la prochaine zone allocable
  // Dans le doute, on demande une page en START_ADDR
  int result;
  int p;
  memheader * current;
  memheader * next;
  
  result = kpage_alloc((void *)START_ADDR, 0);
  if(result == 0)
    {
      // Le Heap tait vide. Donc on crit la structure de dmarrage
      memheader * Header;
      Header = (memheader *)START_ADDR;
      Header->previous = 0x0;
      Header->next = 0x0;
      Header->state = FREE;
    }
  
  if((result < 0x1000)&&(result != 2))
    {
      return NULL;
    }
  // Bon, on a qelque chose de propre, maintenant. On peut commencer  faire tourner l'algo
  // On suppose qu'on a bien plus d'espace addressable que de mmoire, on va donc supposer
  // qu'on atteindra jamais la limite de l'addressage virtuel.
  
  current = (memheader *)START_ADDR;
  
  for(p = 0; p < LIMIT; p++)
    {
      // regardons combien on a de place entre cette structure et la prochaine:
    
      if(current->next == NULL)
	{
	  // On peut s'installer ici. Il faut allouer toute la mmoire, puis remplir correctemnent la structure
	  char * i;
	  for(i = (char *)((unsigned int)current & 0xfffff000); i < (char *)(24 + size + (unsigned int)current); i+= 0x1000)
	    {
	      int Res = kpage_alloc(i, 0);
	      if((Res < 0x1000)&&(Res != 2))
		{
		  // Erreur, on sort
		  return NULL;
		}
	    }
	  // Tout ce qui faut a t allou, on peut continuer!!
	  // Ecrivons la prochaine structure
	  next = (memheader *)(((char *)current) + size + 12);
	  current->next = next;
	  next->previous = current;
	  next->next = 0x0;
	  next->state = FREE;
	  current->state = 0x0;
	  // Fini! On renvoie l'adresse correcte!!
	  return (void *)(((unsigned int)current)+12);
	}
      
      if(current->state & FREE)
	{
	  // Faut voir si on a suffisemment de place
	  int libre = (unsigned int)current->next - (unsigned int)current - 12;
	  if(libre < size)
	    {
	      // On peut s'installer ici. Allouons la mmoire...
	      char * i;
	      for(i = (char *)((unsigned int)current & 0xfffff000); i < (char *)(24 + size + ((unsigned int)current & 0xfffff000)); i+= 0x1000)
		{
		  int Res = kpage_alloc(i, 0);
		  if((Res < 0x1000)&&(Res != 2))
		    {
		      // Erreur, on sort
		      return NULL;
		    }
		}
	      // Tout ce qui faut a t allou, on peut continuer!!
	      // Ecrivons la prochaine structure
	      if(libre - size > 12)
		{
				// On crit une nouvelle structure
		  memheader * inter;
		  inter = (memheader *)(size + (unsigned int)current);
		  inter->previous = current;
		  inter->next = current->next;
		  current->next->previous = inter;
		  current->next = inter;
		  current->state = 0;
		  inter->state = FREE;
				// Et on renvoie l'addresse de la structure
		  return (void *)(((unsigned int)current)+12);
		}
	      else
		{
		  // Pas assez de place, on va plutt tout garder ici
		  current->state = 0x0;
		  return (void *)(((unsigned int)current)+12);
		}
	    }
	}
      // On passe  la prochaine
      current = current->next;
    }
}
  
void kfree(void * p)
{
  // Bon, on va chercher la structure. C'est l qu'on risque de faire un joyeux seg-fault si on se gourre
  memheader * addr;
//  printk("Tentative de libration\n");
  addr = (memheader *)((int)p - 12);
  // Maintenant, le moment de vrit. Est-ce qu'on retrouve des "choses"?
  if(addr->state == 0x0)
    {
      // Bon, on trouve 0, c'est plutt bon signe. Maintenant, on libre
      unsigned int * min; // Min et max librs, pour faire mumuse avec les kpage_free
      unsigned int * max;
      memheader * prec;
      memheader * suiv;
      unsigned int * c;
      
  //    printk("A priori, on a retrouv nos petits\n");

      addr->state = FREE;
      prec = addr->previous;
      suiv = addr->next;

      min = ((unsigned int *)addr) + 12;
      max = ((unsigned int *)suiv);

      if((prec != 0x0)&&(prec->state == FREE))
	{
	  // Dans ce cas, on peut supprimer la structure actuelle. La prcdente devient donc actuelle
	  prec->next = suiv;
	  suiv->previous = prec;
	  addr = prec;
	  min = ((unsigned int *)addr) + 12;
	}
      if(suiv == 0x0)
	{
	  // On est en fin de truc, donc on doit tout effacer
	  addr->next = 0x0;
	  max = max + 0x3000; // Pour tre sr de bien tout effacer
	}
      else
	{
	  if(suiv->state == FREE)
	    {
	      addr->next = suiv->next;
	      max = (unsigned int *)(addr->next);
	    }
	}

      // Maintenant, on peut librer les pages qui ne sont plus utilises
      min = (unsigned int *)(((unsigned int)min / 0x1000) * 0x1000 + 0x1000);
      max = (unsigned int *)(((unsigned int)max / 0x1000) * 0x1000);
//      printk("min = %x - max = %x\n", min, max);
      for(c = min; c < max; c += 1024)
	{
//	  printk("On libre %x\n", c);
	  kpage_free(c, 0);
	}
    }
}


     


