//////////////////
// Gestion de la mmoire
/////////////////

#include <mem.h>
#include <process.h>

/************************************************************/
/* Dfinitions                                              */
/************************************************************/


#define MEM_ZONE 0x4000
#define MEM_ZONE_SIZE 0x1000	// Pour 8 megs de Ram
#define PERM_MEM 0x3000

// Masques

#define ALLOC_MASQUE	0x80

/************************************************************/
/* Prototypes des fonctions prives                         */
/************************************************************/

void * ChercheEmplacementVide();
void SetMemInfos(void ** pagerep, void ** ttpage, void ** acces, int NumProcess);

/************************************************************/
/* Implmentation des fonctions publiques                   */
/************************************************************/


/* Allocation d'une page
Paramtres
	entry -> Zone d'implantation virtuelle de la page
	pagerep -> Adresse virtuelle du rpertoire des pages
  
Codes de retour
	0  Tout va bien
	-1 Plus de place en mmoire relle
	-2 Zone dj occupe

	** Gestion du multiprocessing **
Les paramtres changent! Maintenant, on ne doit plus transfrer pagerep et ttpage, mais
le numro du process. Si le process est 0, il s'agit du kernel, sinon c'est un 
processus utilisateur, qui possde donc son propre triplet de pages:
pagerep, pg0, ttpage.

Voili!!
*/
// OLD: unsigned int kpage_alloc(void * entry, void * pagerep, void * ttpage)
unsigned int kpage_alloc(void * entry, int NumProcess)
{
  void * mem;
  // On va chercher un emplacement vide dans la table des zones, et s'y installer

  mem = ChercheEmplacementVide();
  if(mem == 0) return 1; // Plus de mmoire
  
  return kpage_partage(entry, NumProcess, mem);
}

/* Libre la page situe  l'adresse virtuelle entry
Valeurs de retour:
-1: page non aloue
-2: adresse non valide
-3: bloc systme
-4: adresse de table non valide
-5: bloc de page systme
*/
int kpage_free(void * entry, int NumProcess)
{
  // On choppe l'adresse relle du truc
  unsigned int * p;
  unsigned int * q;
  unsigned int real;
  int result;
  int i;
  void * pagerep;
  void * ttpage;
  void * acces;

  SetMemInfos(&pagerep, &ttpage, &acces, NumProcess);

  p = (unsigned int *)pagerep + (((unsigned int)entry & 0xffc00000)>>22);

  if(*p == 0)
    {
      // La page est dj libre.
      return -1;
    }

  // Sinon, on cherche un cran au dessous
  // On va  l'adresse virtuelle de la table des pages, situ en acces + rent
  p = (unsigned int*)acces + (((unsigned int)entry & 0xffc00000)>>12);
  
  // On dcale de l'adresse correspondante
  p += (((int)entry & 0x003ff000)>>12);

  if(*p == 0)
    {
      // La page est dj libre.
      return -1;
    }

  // Sinon, on choppe l'adresse relle, et on libre
  real = *p;

  result = VideEmplacement(real);

  if(result < 0)
    {
      return result;
    }

  *p = 0;

  // maintenant, il faut librer la table des pages, si elle est vide
  // On va  l'adresse virtuelle de la table des pages, situ en acces + rent
  p = (unsigned int*)pagerep + (((unsigned int)entry & 0xffc00000)>>22);
  q = (unsigned int*)acces + (((unsigned int)entry & 0xffc00000)>>12);

  for(i = 0; i < 1024; i++)
    {
      if(q[i] != 0) break;
    }
  
  if(i >= 1024)
    {
      // On a tout fait, et tout est vide... Librons la table des pages!!
      result = VideEmplacement(*p);

      if(result < 0)
	{
	  return (result - 2);
	}

      // Sinon, on vire effectivement la page
      printk("Table supprime\n");
      *p = 0;
      }
  

  return 0;
}

int kpartage_free(void * entry, int NumProcess)
{
  // On choppe l'adresse relle du truc
  unsigned int * p;
  unsigned int * q;
  unsigned int real;
  int result;
  int i;
  void * pagerep;
  void * ttpage;
  void * acces;


  SetMemInfos(&pagerep, &ttpage, &acces, NumProcess);

  p = (unsigned int *)pagerep + (((unsigned int)entry & 0xffc00000)>>22);

  if(*p == 0)
    {
      // La page est dj libre.
      return -1;
    }


  // Sinon, on cherche un cran au dessous
  p = (unsigned int *)acces + (((unsigned int)entry & 0xffc00000)>>12) + (((unsigned int)entry & 0x003ff000)>>12);

  if(*p == 0)
    {
      // La page est dj libre.
      return -1;
    }

  // Sinon, on choppe l'adresse relle, et on libre surtout pas!!!!
  real = *p;

  if(result < 0)
    {
      return result;
    }

  *p = 0;

  // maintenant, il faut librer la table des pages, si elle est vide
  p = (unsigned int *)pagerep + (((unsigned int)entry & 0xffc00000)>>22);

  // On rcupre l'adresse de dbut de page
  q = (unsigned int *)acces + (((unsigned int)entry & 0xffc00000)>>12);

  for(i = 0; i < 1024; i++)
    {
      if(q[i] != 0) break;
    }
  
 if(i >= 1024)
    {
      // On a tout fait, et tout est vide... Librons la table des pages!!
      result = VideEmplacement(p);

      if(result < 0)
	{
	  return (result - 2);
	}

      // Sinon, on vire effectivement la page
      *p = 0;
      }
  

  return 0;

}

// Old unsigned int kpage_partage(void * entry, void * pagerep, void * ttpage, void * mem)
unsigned int kpage_partage(void * entry, int NumProcess, void * mem)
{
  int * rent;
  int * addrtb;	// Addresse de la table des pages courante
  int * p;		// Pointeur sur l'entre dans la table des pages

  void * pagerep; // Rpertoire des pages
  void * ttpage; // Page des tables
  void * acces; // Adresse virtuelle d'accs aux tables
  

  // Remplissage des infos en fontion du numro de process
  SetMemInfos(&pagerep, &ttpage, &acces, NumProcess);
  
  // On va chercher un emplacement vide dans la table des zones, et s'y installer

  // On vrifie que la table des pages correspondante existe
  rent = (unsigned int*)pagerep + (((unsigned int)entry & 0xffc00000)>>22);
  
  if(*rent == 0)
    {
      // La table des pages n'est pas alloue, il faut donc le faire
      void * table;
      unsigned int * indir;
      int * i;
      int * debut;
      table = ChercheEmplacementVide();
      if(table == 0) 
	{
	  VideEmplacement(mem);
	  return 1;
	}
     
      // Sauvegarde de l'addresse relle dans le rpertoire des pages
      *rent = ((unsigned int)table & 0xfffff000)+7;
      
      // Sauvegarde de l'adresse virtuelle
      indir = (unsigned int *)pagerep + (((unsigned int)entry & 0xffc00000)>>22);
      *indir = *rent;

      // Sauvegarde de la page par un niveau d'indirection en dessous
      indir = (unsigned int *)ttpage + (((unsigned int)entry & 0xffc00000)>>22);
      
      *indir = *rent;

      // Mise  0 de la page
      

      debut = (unsigned int*)acces + (((unsigned int)entry & 0xffc00000)>>12);

      for(i = debut; i<debut + 1024; i++)
	{
	  *i = 0x0;
	}
    }  
  // On va  l'adresse virtuelle de la table des pages, situ en acces + rent
  p = (unsigned int*)acces + (((unsigned int)entry & 0xffc00000)>>12);
  
  // On dcale de l'adresse correspondante
  p += (((int)entry & 0x003ff000)>>12);

  if(*p == 0)
    {
      // Rien! On s'installe
      *p = ((int)mem & 0xFFFFF000)+7;
    }
  else
    {
      VideEmplacement(mem);
      return 2;
    }
  

  return (unsigned int)mem;

}


/************************************************************/
/* Fonctions prives                                        */
/************************************************************/

void * ChercheEmplacementVide()
{
  unsigned char * free;

  for(free = (unsigned char *) MEM_ZONE; 
      free < (unsigned char * )(MEM_ZONE + MEM_ZONE_SIZE); 
      free++)
    {
      if(!((*free) & ALLOC_MASQUE))
	{
	  // On a trouv une zone vide! On la sauve, et on renvoie son adresse
	  *free = 0x80;
	  return (void *)(((unsigned int)free - MEM_ZONE) * 0x1000);
	}
    }
  return 0;
}

int VideEmplacement(int addr)
{
  unsigned char * bloc;
  bloc = (unsigned char *)MEM_ZONE + (addr / 0x1000);

  if((bloc < (unsigned char *)MEM_ZONE)&&(bloc > (unsigned char *)(MEM_ZONE + MEM_ZONE_SIZE)))
    {
      // Erreur! On est en dehors de la zone, on vire
      return -2;
    } 

  if(*bloc & 0x40)
    {
      // Bloc systme, pas touche!!
      return -3;
    }

  // Tout va bien, on relache le bloc
  *bloc = 0x0;

  return 0;
}

void SetMemInfos(void ** pagerep, void ** ttpage, void ** acces, int NumProcess)
{
  if(NumProcess == 0)
    {
      // Processus systme
      *pagerep = (void *)0x1000;
      *ttpage = (void *)0x3000;
      *acces = (void *)0x10000000;
    }
  // TODO: Rcupration des zones pour les processus utilisateurs
  else if(NumProcess >= 100)
    {
      unsigned int prep = (NumProcess - 100)*0x2000 + 0x100000;
      unsigned int tt = (NumProcess - 100)*0x2000 + 0x101000;
      unsigned int acc = 0x70000000;
      unsigned int * p;
      p = (unsigned int *)0x1000 + (0x70000000 >> 22);
      *p = 7+(int) current->meminfo;     //0x103000 + 7;
      *pagerep = (void *)prep;
      *ttpage = (void *)tt;
      *acces = (void *)acc;
    }
  else
    {
      panic("mem - SetMemInfos - Numro de process invalide\n");
    }
}
